UNIT CDPSIDClass;

INTERFACE

USES Classes, SysUtils;

TYPE TCDPSIDGrid = ARRAY[Byte] OF Byte;

     TCDPSIDSystem = FUNCTION ( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer OF OBJECT;

     TCDPSIDState = ( sReady, sOne, sTwo, sThree );

     TCDPSIDClass = CLASS(TObject)
                    PRIVATE { Dclarations prives }

                     FSaveGrid : TCDPSIDGrid;

                     FKeyGrid : TCDPSIDGrid;
                     FSquareGrid : TCDPSIDGrid;
                     FSystems : ARRAY[0..4] OF TCDPSIDSystem;

                     FState : Integer;
                     Fs1, Fs2 : Integer;

                     FUNCTION System1( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
                     FUNCTION System2r( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
                     FUNCTION System2c( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
                     FUNCTION System3r( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
                     FUNCTION System3c( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
                     PROCEDURE Permutation( s, s1 : Integer );
                     PROCEDURE CipherOneByteBuffer( c : Integer; Buffer : PChar; VAR Len : Integer );
                    PROTECTED { Dclarations protges }
                    PUBLIC { Dclarations publiques }
                     CONSTRUCTOR Create;
                     DESTRUCTOR Destroy; OVERRIDE;

                     PROCEDURE LoadFromFile( CONST FileName : STRING );
                     PROCEDURE LoadFromStream( Stream : TStream );

                     PROCEDURE SaveToFile( CONST FileName : STRING );
                     PROCEDURE SaveToStream( Stream : TStream );

                     PROCEDURE CipherToFile( CONST FileNameIn, FileNameOut : STRING );
                     PROCEDURE CipherToStream( StreamIn, StreamOut : TStream );
                     PROCEDURE CipherToBuffer( BufferIn : PChar; LenIn : Integer; BufferOut : PChar; VAR LenOut : Integer );

                     PROCEDURE DeCipherReset;
                     PROCEDURE DeCipherToFile( CONST FileNameIn, FileNameOut : STRING );
                     PROCEDURE DeCipherToStream( StreamIn, StreamOut : TStream );
                     PROCEDURE DeCipherToBuffer( BufferIn : PChar; LenIn : Integer; BufferOut : PChar; VAR LenOut : Integer );

                     PROCEDURE Reset;
                     PROCEDURE Randomize;
                     PROCEDURE Validate;
                    END;

IMPLEMENTATION

// common.h

FUNCTION Square( R, C : Integer ) : Integer;
 BEGIN
  Result:=((R SHL 4)+C) AND 255;
 END; {* FUNC Square *}

FUNCTION Row( Sq : Integer ) : Integer;
 BEGIN
  Result:=(Sq SHR 4) AND 15;
 END; {* FUNC Row *}

FUNCTION Col( Sq : Integer ) : Integer;
 BEGIN
  Result:=Sq AND 15;
 END; {* FUNC Col *}

FUNCTION VAdj2( sa, sb : Integer ) : Boolean;
 VAR d : Integer;
 BEGIN
  IF (Col(sa)<>Col(sb)) THEN
   Result:=False
  ELSE
   BEGIN
    d:=(Row(sa)-Row(sb)) AND 15;
    Result:=(d=1) OR (d=15);
   END;
 END; {* FUNC VAdj2 *}

FUNCTION HAdj2( sa, sb : Integer ) : Boolean;
 VAR d : Integer;
 BEGIN
  IF (Row(sa)<>Row(sb)) THEN
   Result:=False
  ELSE
   BEGIN
//    d:=(Col(sa)-Col(sb)) AND 15;
    d:=(sa-sb) AND 15;
    Result:=(d=1) OR (d=15);
   END;
 END; {* FUNC HAdj2 *}

FUNCTION Adj2( sa, sb : Integer ) : Boolean;
 BEGIN
  Result:=VAdj2(sa, sb) OR HAdj2(sa, sb);
 END; {* FUNC Adj2 *}

FUNCTION VAdj3( sa, sb, sc : Integer ) : Boolean;
 BEGIN
  Result:=VAdj2(sa, sb) AND (VAdj2(sa, sc) OR VAdj2(sb, sc)) OR VAdj2(sa, sc) AND VAdj2(sb, sc);
 END; {* FUNC VAdj3 *}

FUNCTION HAdj3( sa, sb, sc : Integer ) : Boolean;
 BEGIN
  Result:=HAdj2(sa, sb) AND (HAdj2(sa, sc) OR HAdj2(sb, sc)) OR HAdj2(sa, sc) AND HAdj2(sb, sc);
 END; {* FUNC HAdj3 *}

FUNCTION Adj3( sa, sb, sc : Integer ) : Boolean;
 BEGIN
  Result:=VAdj3(sa, sb, sc) OR HAdj3(sa, sb, sc);
 END; {* FUNC Adj3 *}

CONST HexString : ARRAY[0..15] OF Char =  ( '0', '1', '2', '3', '4', '5', '6', '7',
                                            '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' );


FUNCTION ByteToHex( c : Byte ) : STRING;
 BEGIN
  Result:=HexString[c SHR 4]+HexString[c AND 15];
 END;

FUNCTION HexToByte( CONST h : STRING ) : Byte;
 BEGIN
  Result:=16*(Pos(LowerCase(h[1]), HexString)-1)+Pos(LowerCase(h[2]), HexString)-1;
 END;

{ TCDPSIDClass }

PROCEDURE TCDPSIDClass.CipherOneByteBuffer( c : Integer; Buffer : PChar; VAR Len : Integer );
 VAR s : Integer;
     r : Integer;
 BEGIN
  s:=FSQuareGrid[c];
  r:=Random(5);
  Permutation(s, FSystems[r](s, Buffer, Len));
 END; {* PROC .CipherOneByteBuffer *}

PROCEDURE TCDPSIDClass.CipherToBuffer( BufferIn : PChar; LenIn : Integer; BufferOut : PChar; VAR LenOut : Integer );
 VAR Len : Integer;
 BEGIN
  LenOut:=0;
  WHILE (LenIn>0) DO
   BEGIN
    CipherOneByteBuffer(Byte(BufferIn^), BufferOut, Len);
    Inc(LenOut, Len);
    Inc(BufferOut, Len);
    Dec(LenIn);
    Inc(BufferIn);
   END;
 END; {* PROC .CipherToBuffer *}

PROCEDURE TCDPSIDClass.CipherToFile( CONST FileNameIn, FileNameOut : STRING );
 VAR FileIn, FileOut : TFileStream;
 BEGIN
  FileIn:=TFileStream.Create(FileNameIn, fmOpenRead);
  TRY
   FileOut:=TFileStream.Create(FileNameOut, fmCreate);
   TRY
    CipherToStream(FileIn, FileOut);
   FINALLY
    FileOut.Free;
   END;
  FINALLY
   FileIn.Free;
  END;
 END; {* PROC .CipherToFile *}

PROCEDURE TCDPSIDClass.CipherToStream( StreamIn, StreamOut : TStream );
 CONST BufferInSize = 4192;
       BufferOutSize = BufferInSize*3;
 VAR BufferIn, BufferOut : PChar;
     LenIn, LenOut : Integer;
 BEGIN
  GetMem(BufferIn, BufferInSize);
  TRY
   GetMem(BufferOut, BufferOutSize);
   TRY
    LenIn:=StreamIn.Read(BufferIn^, BufferInSize);
    WHILE (LenIn>0) DO
     BEGIN
      CipherToBuffer(BufferIn, LenIn, BufferOut, LenOut);
      StreamOut.Write(BufferOut^, LenOut);
      LenIn:=StreamIn.Read(BufferIn^, BufferInSize);
     END;
   FINALLY
    FreeMem(BufferOut);
   END;
  FINALLY
   FreeMem(BufferIn);
  END;
 END; {* PROC .CipherToStream *}

CONSTRUCTOR TCDPSIDClass.Create;
 VAR i : Integer;
 BEGIN
  INHERITED Create;
  FSystems[0]:=System1;
  FSystems[1]:=System2r;
  FSystems[2]:=System2c;
  FSystems[3]:=System3r;
  FSystems[4]:=System3c;
  DeCipherReset;
  FOR i:=0 TO 255 DO
   BEGIN
    FSaveGrid[i]:=i;
    FKeyGrid[i]:=i;
    FSquareGrid[i]:=i;
   END;
 END; {* CONS .Create *}

PROCEDURE TCDPSIDClass.DeCipherReset;
 BEGIN
  FState:=0;
  Fs1:=0;
  Fs2:=0;
 END; {* PROC .DeCipherReset *}

PROCEDURE TCDPSIDClass.DeCipherToBuffer( BufferIn : PChar; LenIn : Integer; BufferOut : PChar; VAR LenOut : Integer );
 VAR s, s3 : Integer;
 BEGIN
  LenOut:=0;
  WHILE (LenIn>0) DO
   BEGIN
    CASE FState OF
     0 : BEGIN
          Fs1:=FSquareGrid[Byte(BufferIn^)];
          Inc(FState);
         END;
     1 : BEGIN
          Fs2:=FSquareGrid[Byte(BufferIn^)];
          IF NOT Adj2(Fs1, Fs2) THEN
           BEGIN
            s:=Square(Row(Fs1), Col(Fs2));
            Byte(BufferOut^):=FKeyGrid[s];
            Inc(BufferOut);
            Inc(LenOut);
            Permutation(s, Fs1);
            FState:=0;
           END
          ELSE
           Inc(FState);
         END;
     2 : BEGIN
          s3:=FSquareGrid[Byte(BufferIn^)];
          IF NOT Adj3(Fs1, Fs2, s3) THEN
           IF HAdj2(Fs1, Fs2) THEN
             s:=Square(Row(Fs1), Col(s3)) // System2r
            ELSE
             s:=Square(Row(s3), Col(Fs1)) // System2c
           ELSE
            s:=Fs2; // System3r/c
          Byte(BufferOut^):=FKeyGrid[s];
          Inc(BufferOut);
          Inc(LenOut);
          Permutation(s, Fs1);
          FState:=0;
         END;
    END;
    Inc(BufferIn);
    Dec(LenIn);
   END;
 END; {* PROC .DeCipherToBuffer *}

PROCEDURE TCDPSIDClass.DeCipherToFile( CONST FileNameIn, FileNameOut : STRING );
 VAR FileIn, FileOut : TFileStream;
 BEGIN
  FileIn:=TFileStream.Create(FileNameIn, fmOpenRead);
  TRY
   FileOut:=TFileStream.Create(FileNameOut, fmCreate);
   TRY
    DeCipherToStream(FileIn, FileOut);
   FINALLY
    FileOut.Free;
   END;
  FINALLY
   FileIn.Free;
  END;
 END; {* PROC .DeCipherToFile *}

PROCEDURE TCDPSIDClass.DeCipherToStream( StreamIn, StreamOut : TStream );
 CONST BufferOutSize = 4192;
       BufferInSize = BufferOutSize*3;
 VAR BufferIn, BufferOut : PChar;
     LenIn, LenOut : Integer;
 BEGIN
  GetMem(BufferIn, BufferInSize);
  TRY
   GetMem(BufferOut, BufferOutSize);
   TRY
    DeCipherReset;
    LenIn:=StreamIn.Read(BufferIn^, BufferInSize);
    WHILE (LenIn>0) DO
     BEGIN
      DeCipherToBuffer(BufferIn, LenIn, BufferOut, LenOut);
      StreamOut.Write(BufferOut^, LenOut);
      LenIn:=StreamIn.Read(BufferIn^, BufferInSize);
     END;
   FINALLY
    FreeMem(BufferOut);
   END;
  FINALLY
   FreeMem(BufferIn);
  END;
 END; {* PROC .DeCipherToStream *}

DESTRUCTOR TCDPSIDClass.Destroy;
 BEGIN

  INHERITED Destroy;
 END; {* DEST .Destroy *}


PROCEDURE TCDPSIDClass.LoadFromFile( CONST FileName : STRING );
 VAR F : TFileStream;
 BEGIN
  F:=TFileStream.Create(FileName, fmOpenRead);
  TRY
   LoadFromStream(F);
  FINALLY
   F.Free;
  END;
 END; {* PROC .LoadFromFile *}

PROCEDURE TCDPSIDClass.LoadFromStream( Stream : TStream );
 VAR i, j, k : Integer;
     Ch : Char;
 BEGIN
  i:=0;
  j:=0;
  k:=0;
  WHILE (Stream.Read(Ch, 1)=1) AND (i<256) DO
   IF (ch IN ['0'..'9', 'a'..'z', 'A'..'Z']) THEN
    BEGIN
     k:=16*k+Pos(LowerCase(Ch), HexString)-1;
     Inc(j);
     IF (j=2) THEN
      BEGIN
       FSaveGrid[i]:=k;
       FKeyGrid[i]:=k;
       FSquareGrid[k]:=i;
       k:=0;
       j:=0;
       Inc(i);
      END;
    END;
 END; {* PROC .LoadFromStream *}

PROCEDURE TCDPSIDClass.Permutation( s, s1 : Integer );
 VAR ssRow, ssCol, ss : Integer;
     oss, os : Integer;
     rs1, cs1 : Integer;
 BEGIN
  rs1:=Row(s1);
  cs1:=Col(s1);
  // ssRow:=(rs1+Col(s)-cs1)) AND 15;
  ssRow:=(rs1+s-s1) AND 15;
  ssCol:=(cs1+rs1-Row(s)) AND 15;
  ss:=Square(ssRow, ssCol);
  oss:=FKeyGrid[ss];
  os:=FKeyGrid[s];
  FKeyGrid[ss]:=os;
  FKeyGrid[s]:=oss;
  FSquareGrid[oss]:=s;
  FSquareGrid[os]:=ss;
 END; {* PROC .Permutation *}

PROCEDURE TCDPSIDClass.Randomize;
 VAR i, j, k : Integer;
 BEGIN
  FOR i:=0 TO 255 DO
   BEGIN
    FKeyGrid[i]:=i;
    FSquareGrid[i]:=i;
   END;
  FOR i:=0 TO 255 DO
   BEGIN
    j:=Random(255);
    k:=FKeyGrid[i];
    FKeyGrid[i]:=FKeyGrid[j];
    FKeyGrid[j]:=k;

    FSquareGrid[FKeyGrid[i]]:=i;
    FSquareGrid[FKeyGrid[j]]:=j;
   END;

 END; {* PROC .Randomize *}

PROCEDURE TCDPSIDClass.Reset;
 VAR i : Integer;
     k : Integer;
 BEGIN
  FOR i:=0 TO 255 DO
   BEGIN
    k:=FSaveGrid[i];
    FKeyGrid[i]:=k;
    FSquareGrid[k]:=i;
   END;
 END; {* PROC .Reset *}

PROCEDURE TCDPSIDClass.SaveToFile( CONST FileName : STRING );
 VAR F : TFileStream;
 BEGIN
  F:=TFileStream.Create(FileName, fmCreate);
  TRY
   SaveToStream(F);
  FINALLY
   F.Free;
  END;
 END; {* PROC .SaveToFile *}

PROCEDURE TCDPSIDClass.SaveToStream( Stream : TStream );
 VAR r, c : Integer;
     s : STRING;
 BEGIN
  s:='';
  FOR r:=0 TO 15 DO
   BEGIN
    s:=ByteToHex(FSaveGrid[Square(r, 0)]);
    FOR c:=1 TO 15 DO
     s:=s+' '+ByteToHex(FSaveGrid[Square(r, c)]);
    s:=s+#13#10;
    Stream.Write(PChar(s)^, Length(s));
   END;
 END; {* PROC .SaveToStream *}

FUNCTION TCDPSIDClass.System1( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
 VAR s1, s2 : Integer;
     rs, cs : Integer;
 BEGIN
  rs:=Row(s);
  cs:=Col(s);
  REPEAT
   s1:=Square(rs, Random(16));
   s2:=Square(Random(16), cs);
  UNTIL NOT Adj2(s1, s2);
  Byte(Buffer[0]):=FKeyGrid[s1];
  Byte(Buffer[1]):=FKeyGrid[s2];
  Len:=2;
  Result:=s1;
 END; {* FUNC .System1 *}

FUNCTION TCDPSIDClass.System2c( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
 VAR r1, r2, c : Integer;
     s1, s2, s3 : Integer;
     d : Integer;
     rs, cs : Integer;
 BEGIN
  cs:=Col(s);
  rs:=Row(s);
  REPEAT
   c:=Random(16);
   r1:=Random(16);
   d:=Random(2);
   r2:=(r1+2*d-1) AND 15;
   s1:=Square(r1, cs);
   s2:=Square(r2, cs);
   s3:=Square(rs, c);
  UNTIL NOT VAdj3(s1, s2, s3);
  Byte(Buffer[0]):=FKeyGrid[s1];
  Byte(Buffer[1]):=FKeyGrid[s2];
  Byte(Buffer[2]):=FKeyGrid[s3];
  Len:=3;
  Result:=s1;
 ENd; {* FUNC .System2c *}

FUNCTION TCDPSIDClass.System2r( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
 VAR c1, c2, r : Integer;
     s1, s2, s3 : Integer;
     d : Integer;
 BEGIN
  REPEAT
   c1:=Random(16);
   d:=Random(2);
   c2:=(c1+2*d-1) AND 15;
   r:=Random(16);
   s1:=Square(Row(s), c1);
   s2:=Square(Row(s), c2);
   s3:=Square(r, Col(s));
  UNTIL NOT HAdj3(s1, s2, s3);
  Byte(Buffer[0]):=FKeyGrid[s1];
  Byte(Buffer[1]):=FKeyGrid[s2];
  Byte(Buffer[2]):=FKeyGrid[s3];
  Len:=3;
  Result:=s1;
 END; {* FUNC System2r *}

FUNCTION TCDPSIDClass.System3c( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
 VAR ch : Integer;
     s1, s2, s3 : Integer;
     cs, rs : Integer;
 BEGIN
  s2:=s;
  cs:=Col(s);
  rs:=Row(s);

  ch:=Random(4);
  CASE ch OF
   0 : BEGIN
        s1:=Square((rs-1) AND 15, cs);
        s3:=Square((rs+1) AND 15, cs);
       END;
   1 : BEGIN
        s1:=Square((rs-1) AND 15, cs);
        s3:=Square((rs-2) AND 15, cs);
       END;
   2 : BEGIN
        s1:=Square((rs+1) AND 15, cs);
        s3:=Square((rs-1) AND 15, cs);
       END;
   3 : BEGIN
        s1:=Square((rs+1) AND 15, cs);
        s3:=Square((rs+2) AND 15, cs);
       END;
  ELSE
   s1:=0;
   s3:=0;
  END;
  Byte(Buffer[0]):=FKeyGrid[s1];
  Byte(Buffer[1]):=FKeyGrid[s2];
  Byte(Buffer[2]):=FKeyGrid[s3];
  Len:=3;
  Result:=s1;
 END; {* FUNC .System3c *}

FUNCTION TCDPSIDClass.System3r( s : Integer; Buffer : PChar; VAR Len : Integer ) : Integer;
 VAR ch : Integer;
     s1, s2, s3 : Integer;
     rs, cs : Integer;
 BEGIN
  s2:=s;
  rs:=Row(s);
  cs:=Col(s);
  ch:=Random(4);
  CASE ch OF
   0 : BEGIN
        s1:=Square(rs, (cs-1) AND 15);
        s3:=Square(rs, (cs+1) AND 15);
       END;
   1 : BEGIN
        s1:=Square(rs, (cs-1) AND 15);
        s3:=Square(rs, (cs-2) AND 15);
       END;
   2 : BEGIN
        s1:=Square(rs, (cs+1) AND 15);
        s3:=Square(rs, (cs-1) AND 15);
       END;
   3 : BEGIN
        s1:=Square(rs, (cs+1) AND 15);
        s3:=Square(rs, (cs+2) AND 15);
       END;
  ELSE
   s1:=0;
   s3:=0;
  END;
  Byte(Buffer[0]):=FKeyGrid[s1];
  Byte(Buffer[1]):=FKeyGrid[s2];
  Byte(Buffer[2]):=FKeyGrid[s3];
  Len:=3;
  Result:=s1;
 END; {* FUNC .System3r *}

PROCEDURE TCDPSIDClass.Validate;
 VAR i : Integer;
 BEGIN
  FOR i:=0 TO 255 DO
   FSaveGrid[i]:=FKeyGrid[i];
 END; {* PROC .Validate *}

END.
